@@ -34,7 +34,9 @@ using System.Collections.ObjectModel;
 using System.Globalization;
 using System.Linq;
 using System.Reflection;
+#if !FULL_AOT_RUNTIME
 using System.Reflection.Emit;
+#endif
 
 namespace System.Linq.Expressions {
 
@@ -50,11 +52,11 @@ namespace System.Linq.Expressions {
 		internal const BindingFlags AllStatic = BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.FlattenHierarchy;
 		internal const BindingFlags All = BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Instance | BindingFlags.FlattenHierarchy;
 
-		public ExpressionType NodeType {
+		public virtual ExpressionType NodeType {
 			get { return node_type; }
 		}
 
-		public Type Type {
+		public virtual Type Type {
 			get { return type; }
 		}
 
@@ -270,6 +272,12 @@ namespace System.Linq.Expressions {
 
 					if (ltype == rtype && ultype == typeof (bool))
 						return null;
+
+					if (ltype.IsNullable () && ConstantExpression.IsNull (right) && !ConstantExpression.IsNull (left))
+						return null;
+
+					if (rtype.IsNullable () && ConstantExpression.IsNull (left) && !ConstantExpression.IsNull (right))
+						return null;
 				}
 
 				if (oper_name == "op_LeftShift" || oper_name == "op_RightShift") {
@@ -390,12 +398,16 @@ namespace System.Linq.Expressions {
 				if (!left.Type.IsNullable () && !right.Type.IsNullable ()) {
 					is_lifted = false;
 					liftToNull = false;
-					type = typeof (bool);
+					type = typeof(bool);
 				} else if (left.Type.IsNullable () && right.Type.IsNullable ()) {
 					is_lifted = true;
-					type = liftToNull ? typeof (bool?) : typeof (bool);
-				} else
+					type = liftToNull ? typeof(bool?) : typeof(bool);
+				} else if (ConstantExpression.IsNull (left) || ConstantExpression.IsNull (right)) {
+					is_lifted = true;
+					type = typeof (bool);
+				} else {			
 					throw new InvalidOperationException ();
+				}
 			} else {
 				var parameters = method.GetParameters ();
 
@@ -871,6 +883,8 @@ namespace System.Linq.Expressions {
 				return AddChecked (left, right, method);
 			case ExpressionType.AndAlso:
 				return AndAlso (left, right);
+			case ExpressionType.ArrayIndex:
+				return ArrayIndex (left, right);
 			case ExpressionType.Coalesce:
 				return Coalesce (left, right, conversion);
 			case ExpressionType.Divide:
@@ -1043,7 +1057,7 @@ namespace System.Linq.Expressions {
 			if (method == null)
 				return null;
 
-			if (!method.IsGenericMethod && args == null)
+			if (!method.IsGenericMethod && (args == null || args.Length == 0))
 				return method;
 
 			if (args.Length == method.GetGenericArguments ().Length)
@@ -1083,15 +1097,25 @@ namespace System.Linq.Expressions {
 					return false;
 
 				return MethodMatch (closed, name, parameterTypes, argumentTypes);
+			} else if (!method.IsGenericMethod && (argumentTypes != null && argumentTypes.Length > 0))
+				return false;
+
+			for (int i = 0; i < parameters.Length; i++) {
+				var type = parameterTypes [i];
+				var parameter = parameters [i];
+				if (!IsAssignableToParameterType (type, parameter)
+					&& !IsExpressionOfParameter (type, parameter.ParameterType))
+					return false;
 			}
 
-			for (int i = 0; i < parameters.Length; i++)
-				if (!IsAssignableToParameterType (parameterTypes [i], parameters [i]))
-					return false;
-
 			return true;
 		}
 
+		static bool IsExpressionOfParameter (Type type, Type ptype)
+		{
+			return ptype.IsGenericInstanceOf (typeof (Expression<>)) && ptype.GetFirstGenericArgument () == type;
+		}
+
 		static MethodInfo TryGetMethod (Type type, string methodName, BindingFlags flags, Type [] parameterTypes, Type [] argumentTypes)
 		{
 			var methods = from meth in type.GetMethods (flags)
@@ -1099,7 +1123,7 @@ namespace System.Linq.Expressions {
 						  select meth;
 
 			if (methods.Count () > 1)
-				throw new InvalidOperationException ("Too much method candidates");
+				throw new InvalidOperationException ("Too many method candidates");
 
 			var method = TryMakeGeneric (methods.FirstOrDefault (), argumentTypes);
 			if (method != null)
@@ -1171,7 +1195,7 @@ namespace System.Linq.Expressions {
 		static bool IsConvertiblePrimitive (Type type)
 		{
 			var t = type.GetNotNullableType ();
-
+	
 			if (t == typeof (bool))
 				return false;
 
@@ -1209,6 +1233,9 @@ namespace System.Linq.Expressions {
 			if (type.IsInterface || target.IsInterface)
 				return true;
 
+			if (type.IsEnum && target == typeof (Enum))
+				return true;
+
 			if (type.IsValueType || target.IsValueType)
 				return false;
 
@@ -1318,7 +1345,7 @@ namespace System.Linq.Expressions {
 				throw new ArgumentNullException ("addMethod");
 			if (arguments == null)
 				throw new ArgumentNullException ("arguments");
-			if (addMethod.Name.ToLower (CultureInfo.InvariantCulture) != "add")
+			if (addMethod.Name.ToLowerInvariant () != "add")
 				throw new ArgumentException ("addMethod");
 			if (addMethod.IsStatic)
 				throw new ArgumentException ("addMethod must be an instance method", "addMethod");
@@ -1698,7 +1725,7 @@ namespace System.Linq.Expressions {
 			var inits = CheckListInit (newExpression, initializers);
 
 			if (addMethod != null) {
-				if (addMethod.Name.ToLower (CultureInfo.InvariantCulture) != "add")
+				if (addMethod.Name.ToLowerInvariant () != "add")
 					throw new ArgumentException ("addMethod");
 
 				var parameters = addMethod.GetParameters ();
@@ -2087,8 +2114,12 @@ namespace System.Linq.Expressions {
 					throw new ArgumentNullException ("expression");
 				if (!expression.Type.IsAssignableTo (propertyAccessor.DeclaringType))
 					throw new ArgumentException ("expression");
-			} else if (expression != null)
-				throw new ArgumentException ("expression");
+			}
+			//
+			// .NET does not mandate that if the property is static, that the expression must be null
+			// fixes a bug exposed by Orchard's ContentItemRecordAlteration.Alteration
+			// else if (expression != null)
+			//		throw new ArgumentException ("expression");
 
 			var prop = GetAssociatedProperty (propertyAccessor);
 			if (prop == null)
@@ -2249,9 +2280,11 @@ namespace System.Linq.Expressions {
 		// This method must be overwritten by derived classes to
 		// compile the expression
 		//
+#if !FULL_AOT_RUNTIME
 		internal virtual void Emit (EmitContext ec)
 		{
 			throw new NotImplementedException (String.Format ("Emit method is not implemented in expression type {0}", GetType ()));
 		}
+#endif
 	}
 }
