@@ -5,8 +5,13 @@
 #include <mono/metadata/threads.h>
 #include <mono/metadata/reflection.h>
 #include <mono/metadata/mempool.h>
+#include <mono/metadata/class-internals.h>
+#include <mono/metadata/threads-types.h>
 #include <mono/io-layer/io-layer.h>
 #include "mono/utils/mono-compiler.h"
+#include "mono/utils/mono-error.h"
+#include "mono/utils/mono-stack-unwinding.h"
+#include "mono/utils/mono-tls.h"
 
 /* 
  * We should find a better place for this stuff. We can't put it in mono-compiler.h,
@@ -25,7 +30,7 @@
  */
 #ifdef _G_BOOLEAN_EXPR
 #undef _G_BOOLEAN_EXPR
-#define _G_BOOLEAN_EXPR(expr) (gsize)(expr)
+#define _G_BOOLEAN_EXPR(expr) ((gsize)(expr) != 0)
 #endif
 
 #if 1
@@ -78,7 +83,7 @@
 #endif
 
 #define MONO_CHECK_ARG(arg, expr)		G_STMT_START{		  \
-     if (!(expr))							  \
+		if (G_UNLIKELY (!(expr)))							  \
        {								  \
 		MonoException *ex;					  \
 		char *msg = g_strdup_printf ("assertion `%s' failed",	  \
@@ -90,7 +95,7 @@
        };				}G_STMT_END
 
 #define MONO_CHECK_ARG_NULL(arg)	    G_STMT_START{		  \
-     if (arg == NULL)							  \
+		if (G_UNLIKELY (arg == NULL))						  \
        {								  \
 		MonoException *ex;					  \
 		if (arg) {} /* check if the name exists */		  \
@@ -98,10 +103,97 @@
 		mono_raise_exception (ex);				  \
        };				}G_STMT_END
 
-
 /* 16 == default capacity */
 #define mono_stringbuilder_capacity(sb) ((sb)->str ? ((sb)->str->length) : 16)
 
+/* 
+ * Macros which cache the results of lookups locally.
+ * These should be used instead of the original versions, if the __GNUC__
+ * restriction is acceptable.
+ */
+
+#ifdef __GNUC__
+
+/* namespace and name should be a constant */
+/* image must be mscorlib since other assemblies can be unloaded */
+#define mono_class_from_name_cached(image,namespace,name) ({ \
+			static MonoClass *tmp_klass; \
+			if (!tmp_klass) { \
+				g_assert (image == mono_defaults.corlib); \
+				tmp_klass = mono_class_from_name ((image), (namespace), (name)); \
+				g_assert (tmp_klass); \
+			}; \
+			tmp_klass; })
+/* name should be a compile-time constant */
+#define mono_class_get_field_from_name_cached(klass,name) ({ \
+			static MonoClassField *tmp_field; \
+			if (!tmp_field) { \
+				tmp_field = mono_class_get_field_from_name ((klass), (name)); \
+				g_assert (tmp_field); \
+			}; \
+			tmp_field; })
+/* eclass should be a run-time constant */
+#define mono_array_class_get_cached(eclass,rank) ({	\
+			static MonoClass *tmp_klass; \
+			if (!tmp_klass) { \
+				tmp_klass = mono_array_class_get ((eclass), (rank));	\
+				g_assert (tmp_klass); \
+			}; \
+			tmp_klass; })
+/* eclass should be a run-time constant */
+#define mono_array_new_cached(domain, eclass, size) mono_array_new_specific (mono_class_vtable ((domain), mono_array_class_get_cached ((eclass), 1)), (size))
+
+#else
+
+#define mono_class_from_name_cached(image,namespace,name) mono_class_from_name ((image), (namespace), (name))
+#define mono_class_get_field_from_name_cached(klass,name) mono_class_get_field_from_name ((klass), (name))
+#define mono_array_class_get_cached(eclass,rank) mono_array_class_get ((eclass), (rank))
+#define mono_array_new_cached(domain, eclass, size) mono_array_new_specific (mono_class_vtable ((domain), mono_array_class_get_cached ((eclass), 1)), (size))
+
+#endif
+
+#ifdef MONO_BIG_ARRAYS
+typedef uint64_t mono_array_size_t;
+typedef int64_t mono_array_lower_bound_t;
+#define MONO_ARRAY_MAX_INDEX G_MAXINT64
+#define MONO_ARRAY_MAX_SIZE  G_MAXUINT64
+#else
+typedef uint32_t mono_array_size_t;
+typedef int32_t mono_array_lower_bound_t;
+#define MONO_ARRAY_MAX_INDEX ((int32_t) 0x7fffffff)
+#define MONO_ARRAY_MAX_SIZE  ((uint32_t) 0xffffffff)
+#endif
+
+typedef struct {
+	mono_array_size_t length;
+	mono_array_lower_bound_t lower_bound;
+} MonoArrayBounds;
+
+struct _MonoArray {
+	MonoObject obj;
+	/* bounds is NULL for szarrays */
+	MonoArrayBounds *bounds;
+	/* total number of elements of the array */
+	mono_array_size_t max_length; 
+	/* we use double to ensure proper alignment on platforms that need it */
+	double vector [MONO_ZERO_LEN_ARRAY];
+};
+
+struct _MonoString {
+	MonoObject object;
+	int32_t length;
+	mono_unichar2 chars [MONO_ZERO_LEN_ARRAY];
+};
+
+#define mono_object_class(obj) (((MonoObject*)(obj))->vtable->klass)
+#define mono_object_domain(obj) (((MonoObject*)(obj))->vtable->domain)
+
+#define mono_string_chars_fast(s) ((mono_unichar2*)(s)->chars)
+#define mono_string_length_fast(s) ((s)->length)
+
+#define mono_array_length_fast(array) ((array)->max_length)
+#define mono_array_addr_with_size_fast(array,size,index) ( ((char*)(array)->vector) + (size) * (index) )
+
 typedef struct {
 	MonoObject obj;
 	MonoObject *identity;
@@ -153,6 +245,8 @@ struct _MonoException {
 	gint32	    hresult;
 	MonoString *source;
 	MonoObject *_data;
+	MonoObject *captured_traces;
+	MonoArray  *native_trace_ips;
 };
 
 typedef struct {
@@ -170,6 +264,11 @@ typedef struct {
 	MonoString *type_name;
 } MonoTypeLoadException;
 
+typedef struct {
+	MonoException base;
+	MonoObject *wrapped_exception;
+} MonoRuntimeWrappedException;
+
 typedef struct {
 	MonoObject   object;
 	MonoObject  *async_state;
@@ -183,6 +282,7 @@ typedef struct {
 	MonoObject  *async_callback;
 	MonoObject  *execution_context;
 	MonoObject  *original_context;
+	gint64	     add_time;
 } MonoAsyncResult;
 
 typedef struct {
@@ -199,11 +299,17 @@ typedef enum {
 	CallType_OneWay = 3
 } MonoCallType;
 
+/* This corresponds to System.Type */
 struct _MonoReflectionType {
 	MonoObject object;
 	MonoType  *type;
 };
 
+typedef struct {
+	MonoReflectionType type;
+	MonoObject *type_info;
+} MonoReflectionMonoType;
+
 typedef struct {
 	MonoObject  object;
 	MonoReflectionType *class_to_proxy;	
@@ -220,6 +326,7 @@ typedef struct {
 	MonoMarshalByRefObject object;
 	gpointer iunknown;
 	GHashTable* itf_hash;
+	MonoObject *synchronization_context;
 } MonoComObject;
 
 typedef struct {
@@ -259,63 +366,71 @@ typedef struct {
 	MonoString *internal_method_name;
 } MonoStackFrame;
 
-struct _MonoThread {
+typedef enum {
+	MONO_THREAD_FLAG_DONT_MANAGE = 1, // Don't wait for or abort this thread
+	MONO_THREAD_FLAG_NAME_SET = 2, // Thread name set from managed code
+} MonoThreadFlags;
+
+struct _MonoInternalThread {
 	MonoObject  obj;
-	int         lock_thread_id; /* to be used as the pre-shifted thread id in thin locks */
+	volatile int lock_thread_id; /* to be used as the pre-shifted thread id in thin locks. Used for appdomain_ref push/pop */
 	HANDLE	    handle;
 	MonoArray  *cached_culture_info;
-	gpointer    unused1;
-	MonoBoolean threadpool_thread;
 	gunichar2  *name;
 	guint32	    name_len;
 	guint32	    state;
 	MonoException *abort_exc;
-	MonoObject *abort_state;
+	int abort_state_handle;
 	guint64 tid;	/* This is accessed as a gsize in the code (so it can hold a 64bit pointer on systems that need it), but needs to reserve 64 bits of space on all machines as it corresponds to a field in managed code */
 	HANDLE	    start_notify;
 	gpointer stack_ptr;
 	gpointer *static_data;
 	gpointer jit_data;
-	gpointer lock_data;
+	void *thread_info; /*This is MonoThreadInfo*, but to simplify dependencies, let's make it a void* here. */
 	MonoAppContext *current_appcontext;
-	int stack_size;
-	MonoObject *start_obj;
-	GSList *appdomain_refs;
+	MonoException *pending_exception;
+	MonoThread *root_domain_thread;
+	MonoObject *_serialized_principal;
+	int _serialized_principal_version;
+	gpointer appdomain_refs;
 	/* This is modified using atomic ops, so keep it a gint32 */
 	gint32 interruption_requested;
 	gpointer suspend_event;
 	gpointer suspended_event;
 	gpointer resume_event;
 	CRITICAL_SECTION *synch_cs;
-	guint8* serialized_culture_info;
-	guint32 serialized_culture_info_len;
-	guint8* serialized_ui_culture_info;
-	guint32 serialized_ui_culture_info_len;
-	MonoObject *execution_context;
+	MonoBoolean threadpool_thread;
 	MonoBoolean thread_dump_requested;
-	gpointer end_stack; /* This is only used when running in the debugger. */
 	MonoBoolean thread_interrupt_requested;
+	gpointer end_stack; /* This is only used when running in the debugger. */
+	int stack_size;
 	guint8	apartment_state;
 	gint32 critical_region_level;
-	guint32 small_id; /* A small, unique id, used for the hazard pointer table. */
+	gint32 managed_id;
+	guint32 small_id;
 	MonoThreadManageCallback manage_callback;
-	MonoException *pending_exception;
+	gpointer interrupt_on_stop;
+	gsize    flags;
+	gpointer android_tid;
+	gpointer thread_pinning_ref;
+	gint32 ignore_next_signal;
+	MonoMethod *async_invoke_method;
 	/* 
 	 * These fields are used to avoid having to increment corlib versions
-	 * when a new field is added to the unmanaged MonoThread structure.
+	 * when a new field is added to this structure.
+	 * Please synchronize any changes with InternalThread in Thread.cs, i.e. add the
+	 * same field there.
 	 */
+	gpointer unused1;
 	gpointer unused2;
-	gpointer unused3;
-	gpointer unused4;
-	gpointer unused5;
-	gpointer unused6;
 };
 
-typedef struct {
-	MonoString *name;
-	MonoReflectionType *type;
-	MonoObject *value;
-} MonoSerializationEntry;
+struct _MonoThread {
+	MonoObject obj;
+	struct _MonoInternalThread *internal_thread;
+	MonoObject *start_obj;
+	MonoObject *ec_to_set;
+};
 
 typedef struct {
 	guint32 state;
@@ -335,24 +450,21 @@ typedef struct {
 	MonoString *LongTimePattern;
 	MonoString *MonthDayPattern;
 	MonoString *YearMonthPattern;
-	MonoString *FullDateTimePattern;
-	MonoString *RFC1123Pattern;
-	MonoString *SortableDateTimePattern;
-	MonoString *UniversalSortableDateTimePattern;
 	guint32 FirstDayOfWeek;
-	MonoObject *Calendar;
 	guint32 CalendarWeekRule;
 	MonoArray *AbbreviatedDayNames;
 	MonoArray *DayNames;
 	MonoArray *MonthNames;
+	MonoArray *GenitiveMonthNames;
 	MonoArray *AbbreviatedMonthNames;
+	MonoArray *GenitiveAbbreviatedMonthNames;
 	MonoArray *ShortDatePatterns;
 	MonoArray *LongDatePatterns;
 	MonoArray *ShortTimePatterns;
 	MonoArray *LongTimePatterns;
 	MonoArray *MonthDayPatterns;
 	MonoArray *YearMonthPatterns;
-	MonoArray *shortDayNames;
+	MonoArray *ShortestDayNames;
 } MonoDateTimeFormatInfo;
 
 typedef struct 
@@ -403,37 +515,37 @@ typedef struct {
 	MonoBoolean is_read_only;
 	gint32 lcid;
 	gint32 parent_lcid;
-	gint32 specific_lcid;
 	gint32 datetime_index;
 	gint32 number_index;
+	gint32 calendar_type;
 	MonoBoolean use_user_override;
 	MonoNumberFormatInfo *number_format;
 	MonoDateTimeFormatInfo *datetime_format;
 	MonoObject *textinfo;
 	MonoString *name;
-	MonoString *displayname;
 	MonoString *englishname;
 	MonoString *nativename;
 	MonoString *iso3lang;
 	MonoString *iso2lang;
-	MonoString *icu_name;
 	MonoString *win3lang;
 	MonoString *territory;
+	MonoArray *native_calendar_names;
 	MonoCompareInfo *compareinfo;
-	const gint32 *calendar_data;
 	const void* text_info_data;
 } MonoCultureInfo;
 
 typedef struct {
 	MonoObject obj;
-	gint32 region_id;
+	gint32 geo_id;
 	MonoString *iso2name;
 	MonoString *iso3name;
 	MonoString *win3name;
 	MonoString *english_name;
+	MonoString *native_name;
 	MonoString *currency_symbol;
 	MonoString *iso_currency_symbol;
 	MonoString *currency_english_name;
+	MonoString *currency_native_name;
 } MonoRegionInfo;
 
 typedef struct {
@@ -449,6 +561,35 @@ typedef struct {
 	guint32 intType;
 } MonoInterfaceTypeAttribute;
 
+/* 
+ * Callbacks supplied by the runtime and called by the modules in metadata/
+ * This interface is easier to extend than adding a new function type +
+ * a new 'install' function for every callback.
+ */
+typedef struct {
+	gpointer (*create_ftnptr) (MonoDomain *domain, gpointer addr);
+	gpointer (*get_addr_from_ftnptr) (gpointer descr);
+	char*    (*get_runtime_build_info) (void);
+	gpointer (*get_vtable_trampoline) (int slot_index);
+	gpointer (*get_imt_trampoline) (int imt_slot_index);
+	void     (*set_cast_details) (MonoClass *from, MonoClass *to);
+	void     (*debug_log) (int level, MonoString *category, MonoString *message);
+	gboolean (*debug_log_is_enabled) (void);
+} MonoRuntimeCallbacks;
+
+typedef gboolean (*MonoInternalStackWalk) (MonoStackFrameInfo *frame, MonoContext *ctx, gpointer data);
+typedef gboolean (*MonoInternalExceptionFrameWalk) (MonoMethod *method, gpointer ip, size_t native_offset, gboolean managed, gpointer user_data);
+
+typedef struct {
+	void (*mono_walk_stack_with_ctx) (MonoInternalStackWalk func, MonoContext *ctx, MonoUnwindOptions options, void *user_data);
+	void (*mono_walk_stack_with_state) (MonoInternalStackWalk func, MonoThreadUnwindState *state, MonoUnwindOptions options, void *user_data);
+	void (*mono_raise_exception) (MonoException *ex);
+	void (*mono_raise_exception_with_ctx) (MonoException *ex, MonoContext *ctx);
+	gboolean (*mono_exception_walk_trace) (MonoException *ex, MonoInternalExceptionFrameWalk func, gpointer user_data);
+	gboolean (*mono_install_handler_block_guard) (MonoThreadUnwindState *unwind_state);
+} MonoRuntimeExceptionHandlingCallbacks;
+
+
 /* used to free a dynamic method */
 typedef void        (*MonoFreeMethodFunc)	 (MonoDomain *domain, MonoMethod *method);
 
@@ -473,10 +614,6 @@ void
 mono_message_init	    (MonoDomain *domain, MonoMethodMessage *this_obj, 
 			     MonoReflectionMethod *method, MonoArray *out_args) MONO_INTERNAL;
 
-MonoObject *
-mono_remoting_invoke	    (MonoObject *real_proxy, MonoMethodMessage *msg, 
-			     MonoObject **exc, MonoArray **out_args) MONO_INTERNAL;
-
 MonoObject *
 mono_message_invoke	    (MonoObject *target, MonoMethodMessage *msg, 
 			     MonoObject **exc, MonoArray **out_args) MONO_INTERNAL;
@@ -500,12 +637,6 @@ mono_class_get_allocation_ftn (MonoVTable *vtable, gboolean for_box, gboolean *p
 void
 mono_runtime_free_method    (MonoDomain *domain, MonoMethod *method) MONO_INTERNAL;
 
-/* runtime initialization functions */
-typedef void (*MonoExceptionFunc) (MonoException *ex);
-
-void
-mono_install_handler	    (MonoExceptionFunc func) MONO_INTERNAL;
-
 void	    
 mono_install_runtime_invoke (MonoInvokeFunc func) MONO_INTERNAL;
 
@@ -515,19 +646,37 @@ mono_install_compile_method (MonoCompileFunc func) MONO_INTERNAL;
 void
 mono_install_free_method    (MonoFreeMethodFunc func) MONO_INTERNAL;
 
+void
+mono_install_callbacks      (MonoRuntimeCallbacks *cbs) MONO_INTERNAL;
+
+MonoRuntimeCallbacks*
+mono_get_runtime_callbacks (void) MONO_INTERNAL;
+
+void
+mono_install_eh_callbacks (MonoRuntimeExceptionHandlingCallbacks *cbs) MONO_INTERNAL;
+
+MonoRuntimeExceptionHandlingCallbacks *
+mono_get_eh_callbacks (void) MONO_INTERNAL;
+
+void
+mono_raise_exception_with_context (MonoException *ex, MonoContext *ctx) MONO_INTERNAL;
+
 void
 mono_type_initialization_init (void) MONO_INTERNAL;
 
 void
 mono_type_initialization_cleanup (void) MONO_INTERNAL;
 
-guint32
+int
+mono_thread_kill           (MonoInternalThread *thread, int signal) MONO_INTERNAL;
+
+MonoNativeTlsKey
 mono_thread_get_tls_key    (void) MONO_INTERNAL;
 
 gint32
 mono_thread_get_tls_offset (void) MONO_INTERNAL;
 
-guint32
+MonoNativeTlsKey
 mono_domain_get_tls_key    (void) MONO_INTERNAL;
 
 gint32
@@ -535,6 +684,43 @@ mono_domain_get_tls_offset (void) MONO_INTERNAL;
 
 /* Reflection and Reflection.Emit support */
 
+/*
+ * Handling System.Type objects:
+ *
+ *   Fields defined as System.Type in managed code should be defined as MonoObject* 
+ * in unmanaged structures, and the monotype_cast () function should be used for 
+ * casting them to MonoReflectionType* to avoid crashes/security issues when 
+ * encountering instances of user defined subclasses of System.Type.
+ */
+
+#define IS_MONOTYPE(obj) (!(obj) || (((MonoObject*)(obj))->vtable->klass->image == mono_defaults.corlib && ((MonoReflectionType*)(obj))->type != NULL))
+
+/* 
+ * Make sure the argument, which should be a System.Type is a System.MonoType object 
+ * or equivalent, and not an instance of 
+ * a user defined subclass of System.Type. This should be used in places were throwing
+ * an exception is safe.
+ */
+#define CHECK_MONOTYPE(obj) do { \
+	if (!IS_MONOTYPE (obj)) \
+		mono_raise_exception (mono_get_exception_not_supported ("User defined subclasses of System.Type are not yet supported")); \
+	} while (0)
+
+/* This should be used for accessing members of Type[] arrays */
+#define mono_type_array_get(arr,index) monotype_cast (mono_array_get ((arr), gpointer, (index)))
+
+/*
+ * Cast an object to MonoReflectionType, making sure it is a System.MonoType or
+ * a subclass of it.
+ */
+static inline MonoReflectionType*
+monotype_cast (MonoObject *obj)
+{
+	g_assert (IS_MONOTYPE (obj));
+
+	return (MonoReflectionType*)obj;
+}
+
 /*
  * The following structure must match the C# implementation in our corlib.
  */
@@ -591,11 +777,17 @@ struct _MonoReflectionProperty {
 	MonoProperty *property;
 };
 
+/*This is System.EventInfo*/
 struct _MonoReflectionEvent {
 	MonoObject object;
+	MonoObject *cached_add_event;
+};
+
+typedef struct {
+	MonoReflectionEvent object;
 	MonoClass *klass;
 	MonoEvent *event;
-};
+} MonoReflectionMonoEvent;
 
 typedef struct {
 	MonoObject object;
@@ -650,6 +842,7 @@ typedef struct {
 
 typedef struct {
 	MonoReflectionType *parent;
+	MonoReflectionType *declaring_type;
 	MonoString *name;
 	MonoReflectionMethod *get;
 	MonoReflectionMethod *set;
@@ -702,7 +895,7 @@ typedef struct {
 } MonoILExceptionInfo;
 
 typedef struct {
-	MonoReflectionType *extype;
+	MonoObject *extype;
 	gint32 type;
 	gint32 start;
 	gint32 len;
@@ -711,7 +904,7 @@ typedef struct {
 
 typedef struct {
 	MonoObject object;
-	MonoReflectionType *catch_type;
+	MonoObject *catch_type;
 	gint32 filter_offset;
 	gint32 flags;
 	gint32 try_offset;
@@ -733,7 +926,7 @@ typedef struct {
 	 * LocalBuilder inherits from it under net 2.0.
 	 */
 	MonoObject object;
-	MonoReflectionType *type;
+	MonoObject *type;
 	MonoBoolean is_pinned;
 	guint16 local_index;
 	MonoString *name;
@@ -747,7 +940,7 @@ typedef struct {
 	MonoString *guid;
 	MonoString *mcookie;
 	MonoString *marshaltype;
-	MonoReflectionType *marshaltyperef;
+	MonoObject *marshaltyperef;
 	gint32 param_num;
 	MonoBoolean has_size;
 } MonoReflectionMarshal;
@@ -785,7 +978,7 @@ typedef struct {
 typedef struct {
 	MonoObject object;
 	MonoMethod *mhandle;
-	MonoReflectionType *rtype;
+	MonoObject *rtype;
 	MonoArray *parameters;
 	guint32 attrs;
 	guint32 iattrs;
@@ -796,7 +989,7 @@ typedef struct {
 	MonoObject *type;
 	MonoArray *pinfo;
 	MonoArray *cattrs;
-	MonoReflectionMethod *override_method;
+	MonoArray *override_methods;
 	MonoString *dll;
 	MonoString *dllentry;
 	guint32 charset;
@@ -872,12 +1065,13 @@ typedef struct {
 	gint32 machine;
 	MonoBoolean corlib_internal;
 	MonoArray *type_forwarders;
+	MonoArray *pktoken; /* as hexadecimal byte[] */
 } MonoReflectionAssemblyBuilder;
 
 typedef struct {
 	MonoObject object;
 	guint32 attrs;
-	MonoReflectionType *type;
+	MonoObject *type;
 	MonoString *name;
 	MonoObject *def_value;
 	gint32 offset;
@@ -895,13 +1089,19 @@ typedef struct {
 	MonoObject object;
 	guint32 attrs;
 	MonoString *name;
-	MonoReflectionType *type;
+	MonoObject *type;
 	MonoArray *parameters;
 	MonoArray *cattrs;
 	MonoObject *def_value;
 	MonoReflectionMethodBuilder *set_method;
 	MonoReflectionMethodBuilder *get_method;
 	gint32 table_idx;
+	MonoObject *type_builder;
+	MonoArray *returnModReq;
+	MonoArray *returnModOpt;
+	MonoArray *paramModReq;
+	MonoArray *paramModOpt;
+	guint32 call_conv;
 } MonoReflectionPropertyBuilder;
 
 struct _MonoReflectionModule {
@@ -934,7 +1134,7 @@ typedef struct {
 	MonoReflectionType type;
 	MonoString *name;
 	MonoString *nspace;
-	MonoReflectionType *parent;
+	MonoObject *parent;
 	MonoReflectionType *nesting_type;
 	MonoArray *interfaces;
 	gint32     num_methods;
@@ -957,6 +1157,17 @@ typedef struct {
 	MonoReflectionType *created;
 } MonoReflectionTypeBuilder;
 
+typedef struct {
+	MonoReflectionType type;
+	MonoReflectionType *element_type;
+	int rank;
+} MonoReflectionArrayType;
+
+typedef struct {
+	MonoReflectionType type;
+	MonoReflectionType *element_type;
+} MonoReflectionDerivedType;
+
 typedef struct {
 	MonoReflectionType type;
 	MonoReflectionTypeBuilder *tbuilder;
@@ -972,7 +1183,8 @@ typedef struct {
 typedef struct _MonoReflectionGenericClass MonoReflectionGenericClass;
 struct _MonoReflectionGenericClass {
 	MonoReflectionType type;
-	MonoReflectionTypeBuilder *generic_type;
+	MonoReflectionType *generic_type; /*Can be either a MonoType or a TypeBuilder*/
+	MonoArray *type_arguments;
 	guint32 initialized;
 };
 
@@ -1012,6 +1224,21 @@ typedef struct {
 	MonoArray *data;
 } MonoReflectionCustomAttr;
 
+typedef struct {
+	MonoObject object;
+	MonoString *marshal_cookie;
+	MonoString *marshal_type;
+	MonoReflectionType *marshal_type_ref;
+	MonoReflectionType *marshal_safe_array_user_defined_subtype;
+	guint32 utype;
+	guint32 array_subtype;
+	gint32 safe_array_subtype;
+	gint32 size_const;
+	gint32 IidParameterIndex;
+	gint16 size_param_index;
+} MonoReflectionMarshalAsAttribute;
+
+
 typedef struct {
 	MonoObject object;
 	gint32 call_conv;
@@ -1072,19 +1299,21 @@ typedef struct {
 typedef struct {
 	MonoObject object;
 	MonoReflectionGenericClass *inst;
-	MonoReflectionFieldBuilder *fb;
+	MonoObject *fb; /*can be either a MonoField or a FieldBuilder*/
 } MonoReflectionFieldOnTypeBuilderInst;
 
 typedef struct {
 	MonoObject object;
 	MonoReflectionGenericClass *inst;
-	MonoReflectionCtorBuilder *cb;
+	MonoObject *cb; /*can be either a MonoCMethod or ConstructorBuilder*/
 } MonoReflectionCtorOnTypeBuilderInst;
 
 typedef struct {
 	MonoObject object;
-	MonoReflectionGenericClass *inst;
-	MonoReflectionMethodBuilder *mb;
+	MonoReflectionType *inst;
+	MonoObject *mb; /*can be either a MonoMethod or MethodBuilder*/
+	MonoArray *method_args;
+	MonoReflectionMethodBuilder *generic_method_definition;
 } MonoReflectionMethodOnTypeBuilderInst;
 
 typedef struct {
@@ -1105,6 +1334,12 @@ typedef struct {
 	guint32 location;
 } MonoManifestResourceInfo;
 
+/* A boxed IntPtr */
+typedef struct {
+	MonoObject object;
+	gpointer m_value;
+} MonoIntPtr;
+
 /* Keep in sync with System.GenericParameterAttributes */
 typedef enum {
 	GENERIC_PARAMETER_ATTRIBUTE_NON_VARIANT		= 0,
@@ -1119,6 +1354,12 @@ typedef enum {
 	GENERIC_PARAMETER_ATTRIBUTE_SPECIAL_CONSTRAINTS_MASK	= 28
 } GenericParameterAttributes;
 
+typedef struct {
+	MonoType *type;
+	MonoClassField *field;
+	MonoProperty *prop;
+} CattrNamedArg;
+
 void          mono_image_create_pefile (MonoReflectionModuleBuilder *module, HANDLE file) MONO_INTERNAL;
 void          mono_image_basic_init (MonoReflectionAssemblyBuilder *assembly) MONO_INTERNAL;
 MonoReflectionModule * mono_image_load_module_dynamic (MonoReflectionAssemblyBuilder *assembly, MonoString *file_name) MONO_INTERNAL;
@@ -1128,6 +1369,8 @@ guint32       mono_image_create_method_token (MonoDynamicImage *assembly, MonoOb
 void          mono_image_module_basic_init (MonoReflectionModuleBuilder *module) MONO_INTERNAL;
 void          mono_image_register_token (MonoDynamicImage *assembly, guint32 token, MonoObject *obj) MONO_INTERNAL;
 void          mono_dynamic_image_free (MonoDynamicImage *image) MONO_INTERNAL;
+void          mono_image_set_wrappers_type (MonoReflectionModuleBuilder *mb, MonoReflectionType *type) MONO_INTERNAL;
+void          mono_dynamic_image_release_gc_roots (MonoDynamicImage *image) MONO_INTERNAL;
 
 void        mono_reflection_setup_internal_class  (MonoReflectionTypeBuilder *tb) MONO_INTERNAL;
 
@@ -1145,6 +1388,13 @@ void mono_reflection_create_dynamic_method (MonoReflectionDynamicMethod *m) MONO
 void mono_reflection_destroy_dynamic_method (MonoReflectionDynamicMethod *mb) MONO_INTERNAL;
 
 void        mono_reflection_initialize_generic_parameter (MonoReflectionGenericParam *gparam) MONO_INTERNAL;
+void        mono_reflection_create_unmanaged_type (MonoReflectionType *type) MONO_INTERNAL;
+void        mono_reflection_register_with_runtime (MonoReflectionType *type) MONO_INTERNAL;
+
+void        mono_reflection_create_custom_attr_data_args (MonoImage *image, MonoMethod *method, const guchar *data, guint32 len, MonoArray **typed_args, MonoArray **named_args, CattrNamedArg **named_arg_info, MonoError *error) MONO_INTERNAL;
+MonoMethodSignature * mono_reflection_lookup_signature (MonoImage *image, MonoMethod *method, guint32 token) MONO_INTERNAL;
+
+MonoArray* mono_param_get_objects_internal  (MonoDomain *domain, MonoMethod *method, MonoClass *refclass) MONO_INTERNAL;
 
 MonoClass*
 mono_class_bind_generic_parameters (MonoClass *klass, int type_argc, MonoType **types, gboolean is_dynamic) MONO_INTERNAL;
@@ -1153,7 +1403,7 @@ mono_reflection_bind_generic_parameters (MonoReflectionType *type, int type_argc
 MonoReflectionMethod*
 mono_reflection_bind_generic_method_parameters (MonoReflectionMethod *method, MonoArray *types) MONO_INTERNAL;
 void
-mono_reflection_generic_class_initialize (MonoReflectionGenericClass *type, MonoArray *methods, MonoArray *ctors, MonoArray *fields, MonoArray *properties, MonoArray *events) MONO_INTERNAL;
+mono_reflection_generic_class_initialize (MonoReflectionGenericClass *type, MonoArray *fields) MONO_INTERNAL;
 MonoReflectionEvent *
 mono_reflection_event_builder_get_event_info (MonoReflectionTypeBuilder *tb, MonoReflectionEventBuilder *eb) MONO_INTERNAL;
 
@@ -1161,7 +1411,7 @@ MonoArray  *mono_reflection_sighelper_get_signature_local (MonoReflectionSigHelp
 
 MonoArray  *mono_reflection_sighelper_get_signature_field (MonoReflectionSigHelper *sig) MONO_INTERNAL;
 
-MonoReflectionMarshal* mono_reflection_marshal_from_marshal_spec (MonoDomain *domain, MonoClass *klass, MonoMarshalSpec *spec) MONO_INTERNAL;
+MonoReflectionMarshalAsAttribute* mono_reflection_marshal_as_attribute_from_marshal_spec (MonoDomain *domain, MonoClass *klass, MonoMarshalSpec *spec) MONO_INTERNAL;
 
 gpointer
 mono_reflection_lookup_dynamic_token (MonoImage *image, guint32 token, gboolean valid_token, MonoClass **handle_class, MonoGenericContext *context) MONO_INTERNAL;
@@ -1172,6 +1422,15 @@ mono_reflection_call_is_assignable_to (MonoClass *klass, MonoClass *oklass) MONO
 gboolean
 mono_reflection_is_valid_dynamic_token (MonoDynamicImage *image, guint32 token) MONO_INTERNAL;
 
+void
+mono_reflection_resolve_custom_attribute_data (MonoReflectionMethod *method, MonoReflectionAssembly *assembly, gpointer data, guint32 data_length, MonoArray **ctor_args, MonoArray ** named_args) MONO_INTERNAL;
+
+MonoType*
+mono_reflection_type_get_handle (MonoReflectionType *ref) MONO_INTERNAL;
+
+void
+mono_reflection_free_dynamic_generic_class (MonoGenericClass *gclass) MONO_INTERNAL;
+
 void
 mono_image_build_metadata (MonoReflectionModuleBuilder *module) MONO_INTERNAL;
 
@@ -1179,10 +1438,14 @@ int
 mono_get_constant_value_from_blob (MonoDomain* domain, MonoTypeEnum type, const char *blob, void *value) MONO_INTERNAL;
 
 void
-mono_release_type_locks (MonoThread *thread) MONO_INTERNAL;
+mono_release_type_locks (MonoInternalThread *thread) MONO_INTERNAL;
 
 char *
-mono_string_to_utf8_mp	(MonoMemPool *mp, MonoString *s) MONO_INTERNAL;
+mono_string_to_utf8_mp	(MonoMemPool *mp, MonoString *s, MonoError *error) MONO_INTERNAL;
+
+char *
+mono_string_to_utf8_image (MonoImage *image, MonoString *s, MonoError *error) MONO_INTERNAL;
+
 
 MonoArray*
 mono_array_clone_in_domain (MonoDomain *domain, MonoArray *array) MONO_INTERNAL;
@@ -1190,11 +1453,20 @@ mono_array_clone_in_domain (MonoDomain *domain, MonoArray *array) MONO_INTERNAL;
 void
 mono_array_full_copy (MonoArray *src, MonoArray *dest) MONO_INTERNAL;
 
+gboolean
+mono_array_calc_byte_len (MonoClass *class, uintptr_t len, uintptr_t *res) MONO_INTERNAL;
+
+#ifndef DISABLE_REMOTING
+MonoObject *
+mono_remoting_invoke	    (MonoObject *real_proxy, MonoMethodMessage *msg, 
+			     MonoObject **exc, MonoArray **out_args) MONO_INTERNAL;
+
 gpointer
 mono_remote_class_vtable (MonoDomain *domain, MonoRemoteClass *remote_class, MonoRealProxy *real_proxy) MONO_INTERNAL;
 
 void
 mono_upgrade_remote_class (MonoDomain *domain, MonoObject *tproxy, MonoClass *klass) MONO_INTERNAL;
+#endif
 
 gpointer
 mono_create_ftnptr (MonoDomain *domain, gpointer addr) MONO_INTERNAL;
@@ -1208,7 +1480,11 @@ mono_nullable_init (guint8 *buf, MonoObject *value, MonoClass *klass) MONO_INTER
 MonoObject*
 mono_nullable_box (guint8 *buf, MonoClass *klass) MONO_INTERNAL;
 
+#ifdef MONO_SMALL_CONFIG
+#define MONO_IMT_SIZE 9
+#else
 #define MONO_IMT_SIZE 19
+#endif
 
 typedef union {
 	int vtable_slot;
@@ -1220,6 +1496,7 @@ typedef struct _MonoImtBuilderEntry {
 	struct _MonoImtBuilderEntry *next;
 	MonoImtItemValue value;
 	int children;
+	guint8 has_target_code : 1;
 } MonoImtBuilderEntry;
 
 typedef struct _MonoIMTCheckItem MonoIMTCheckItem;
@@ -1234,6 +1511,7 @@ struct _MonoIMTCheckItem {
 	guint8            compare_done;
 	guint8            chunk_size;
 	guint8            short_branch;
+	guint8            has_target_code;
 };
 
 typedef gpointer (*MonoImtThunkBuilder) (MonoVTable *vtable, MonoDomain *domain,
@@ -1242,12 +1520,6 @@ typedef gpointer (*MonoImtThunkBuilder) (MonoVTable *vtable, MonoDomain *domain,
 void
 mono_install_imt_thunk_builder (MonoImtThunkBuilder func) MONO_INTERNAL;
 
-void
-mono_install_imt_trampoline (gpointer tramp) MONO_INTERNAL;
-
-void
-mono_install_vtable_trampoline (gpointer tramp) MONO_INTERNAL;
-
 void
 mono_vtable_build_imt_slot (MonoVTable* vtable, int imt_slot) MONO_INTERNAL;
 
@@ -1255,8 +1527,9 @@ guint32
 mono_method_get_imt_slot (MonoMethod *method) MONO_INTERNAL;
 
 void
-mono_method_add_generic_virtual_invocation (MonoDomain *domain, gpointer *vtable_slot,
-	MonoGenericInst *method_inst, gpointer code) MONO_INTERNAL;
+mono_method_add_generic_virtual_invocation (MonoDomain *domain, MonoVTable *vtable,
+											gpointer *vtable_slot,
+											MonoMethod *method, gpointer code) MONO_INTERNAL;
 
 gpointer
 mono_method_alloc_generic_virtual_thunk (MonoDomain *domain, int size) MONO_INTERNAL;
@@ -1280,7 +1553,61 @@ mono_runtime_class_init_full (MonoVTable *vtable, gboolean raise_exception) MONO
 void
 mono_method_clear_object (MonoDomain *domain, MonoMethod *method) MONO_INTERNAL;
 
+void
+mono_class_compute_gc_descriptor (MonoClass *class) MONO_INTERNAL;
+
+gsize*
+mono_class_compute_bitmap (MonoClass *class, gsize *bitmap, int size, int offset, int *max_set, gboolean static_fields) MONO_INTERNAL;
+
+MonoObject*
+mono_object_xdomain_representation (MonoObject *obj, MonoDomain *target_domain, MonoObject **exc) MONO_INTERNAL;
+
+gboolean
+mono_class_is_reflection_method_or_constructor (MonoClass *class) MONO_INTERNAL;
+
+MonoObject *
+mono_get_object_from_blob (MonoDomain *domain, MonoType *type, const char *blob) MONO_INTERNAL;
+
+gpointer
+mono_class_get_ref_info (MonoClass *klass) MONO_INTERNAL;
+
+void
+mono_class_set_ref_info (MonoClass *klass, gpointer obj) MONO_INTERNAL;
+
+void
+mono_class_free_ref_info (MonoClass *klass) MONO_INTERNAL;
+
+MonoObject *
+mono_object_new_pinned (MonoDomain *domain, MonoClass *klass) MONO_INTERNAL;
+
+void
+mono_field_static_get_value_for_thread (MonoInternalThread *thread, MonoVTable *vt, MonoClassField *field, void *value) MONO_INTERNAL;
+
+/* exported, used by the debugger */
+MONO_API void *
+mono_vtable_get_static_field_data (MonoVTable *vt);
+
+char *
+mono_string_to_utf8_ignore (MonoString *s) MONO_INTERNAL;
+
+char *
+mono_string_to_utf8_image_ignore (MonoImage *image, MonoString *s) MONO_INTERNAL;
+
+char *
+mono_string_to_utf8_mp_ignore (MonoMemPool *mp, MonoString *s) MONO_INTERNAL;
+
+gboolean
+mono_monitor_is_il_fastpath_wrapper (MonoMethod *method) MONO_INTERNAL;
+
+char *
+mono_exception_get_native_backtrace (MonoException *exc) MONO_INTERNAL;
+
+MonoString *
+ves_icall_Mono_Runtime_GetNativeStackTrace (MonoException *exc) MONO_INTERNAL;
+
+char *
+mono_exception_get_managed_backtrace (MonoException *exc) MONO_INTERNAL;
+
 #endif /* __MONO_OBJECT_INTERNALS_H__ */
 
 
-
