@@ -9,6 +9,13 @@
  */
 
 #include <config.h>
+
+#if _WIN32_WINNT < 0x0501
+/* Required for Vectored Exception Handling. */
+#undef _WIN32_WINNT
+#define _WIN32_WINNT 0x0501
+#endif /* _WIN32_WINNT < 0x0501 */
+
 #include <glib.h>
 #include <signal.h>
 #include <string.h>
@@ -40,24 +47,37 @@ static MonoW32ExceptionHandler ill_handler;
 static MonoW32ExceptionHandler segv_handler;
 
 LPTOP_LEVEL_EXCEPTION_FILTER mono_old_win_toplevel_exception_filter;
-guint64 mono_win_chained_exception_filter_result;
-gboolean mono_win_chained_exception_filter_didrun;
+guint64 mono_win_vectored_exception_handle;
+extern gboolean mono_win_chained_exception_needs_run;
 
 #define W32_SEH_HANDLE_EX(_ex) \
 	if (_ex##_handler) _ex##_handler(0, ep, sctx)
 
+LONG CALLBACK seh_unhandled_exception_filter(EXCEPTION_POINTERS* ep)
+{
+#ifndef MONO_CROSS_COMPILE
+	if (mono_old_win_toplevel_exception_filter) {
+		return (*mono_old_win_toplevel_exception_filter)(ep);
+	}
+#endif
+
+	mono_handle_native_sigsegv (SIGSEGV, NULL);
+
+	return EXCEPTION_CONTINUE_SEARCH;
+}
+
 /*
  * Unhandled Exception Filter
  * Top-level per-process exception handler.
  */
-LONG CALLBACK seh_handler(EXCEPTION_POINTERS* ep)
+LONG CALLBACK seh_vectored_exception_handler(EXCEPTION_POINTERS* ep)
 {
 	EXCEPTION_RECORD* er;
 	CONTEXT* ctx;
 	MonoContext* sctx;
 	LONG res;
 
-	mono_win_chained_exception_filter_didrun = FALSE;
+	mono_win_chained_exception_needs_run = FALSE;
 	res = EXCEPTION_CONTINUE_EXECUTION;
 
 	er = ep->ExceptionRecord;
@@ -98,40 +118,57 @@ LONG CALLBACK seh_handler(EXCEPTION_POINTERS* ep)
 		break;
 	}
 
-	/* Copy context back */
-	/* Nonvolatile */
-	ctx->Rsp = sctx->rsp; 
-	ctx->Rdi = sctx->rdi; 
-	ctx->Rsi = sctx->rsi; 
-	ctx->Rbx = sctx->rbx; 
-	ctx->Rbp = sctx->rbp;
-	ctx->R12 = sctx->r12; 
-	ctx->R13 = sctx->r13; 
-	ctx->R14 = sctx->r14;
-	ctx->R15 = sctx->r15;
-	ctx->Rip = sctx->rip; 
+	if (win32_chained_exception_needs_run) {
+		/* Don't copy context back if we chained exception
+		* as the handler may have modfied the EXCEPTION_POINTERS
+		* directly. We don't pass sigcontext to chained handlers.
+		* Return continue search so the UnhandledExceptionFilter
+		* can correctly chain the exception.
+		*/
+		res = EXCEPTION_CONTINUE_SEARCH;
+	} else {
+		/* Copy context back */
+		/* Nonvolatile */
+		ctx->Rsp = sctx->rsp;
+		ctx->Rdi = sctx->rdi;
+		ctx->Rsi = sctx->rsi;
+		ctx->Rbx = sctx->rbx;
+		ctx->Rbp = sctx->rbp;
+		ctx->R12 = sctx->r12;
+		ctx->R13 = sctx->r13;
+		ctx->R14 = sctx->r14;
+		ctx->R15 = sctx->r15;
+		ctx->Rip = sctx->rip;
 
-	/* Volatile But should not matter?*/
-	ctx->Rax = sctx->rax; 
-	ctx->Rcx = sctx->rcx; 
-	ctx->Rdx = sctx->rdx;
+		/* Volatile But should not matter?*/
+		ctx->Rax = sctx->rax;
+		ctx->Rcx = sctx->rcx;
+		ctx->Rdx = sctx->rdx;
+	}
 
-	g_free (sctx);
-
-	if (mono_win_chained_exception_filter_didrun)
-		res = mono_win_chained_exception_filter_result;
+	/* TODO: Find right place to free this in stack overflow case */
+	if (er->ExceptionCode != EXCEPTION_STACK_OVERFLOW)
+		g_free (sctx);
 
 	return res;
 }
 
 void win32_seh_init()
 {
-	mono_old_win_toplevel_exception_filter = SetUnhandledExceptionFilter(seh_handler);
+	mono_old_win_toplevel_exception_filter = SetUnhandledExceptionFilter(seh_unhandled_exception_filter);
+	mono_win_vectored_exception_handle = AddVectoredExceptionHandler (1, seh_vectored_exception_handler);
 }
 
 void win32_seh_cleanup()
 {
 	if (mono_old_win_toplevel_exception_filter) SetUnhandledExceptionFilter(mono_old_win_toplevel_exception_filter);
+
+	guint32 ret = 0;
+	if (old_win32_toplevel_exception_filter)
+		SetUnhandledExceptionFilter(mono_old_win_toplevel_exception_filter);
+
+	ret = RemoveVectoredExceptionHandler (mono_win_vectored_exception_handle);
+	g_assert (ret);
 }
 
 void win32_seh_set_handler(int type, MonoW32ExceptionHandler handler)
@@ -209,7 +246,7 @@ mono_arch_get_restore_context (MonoTrampInfo **info, gboolean aot)
 	mono_arch_flush_icache (start, code - start);
 
 	if (info)
-		*info = mono_tramp_info_create (g_strdup_printf ("restore_context"), start, code - start, ji, unwind_ops);
+		*info = mono_tramp_info_create ("restore_context", start, code - start, ji, unwind_ops);
 
 	return start;
 }
@@ -296,7 +333,7 @@ mono_arch_get_call_filter (MonoTrampInfo **info, gboolean aot)
 	mono_arch_flush_icache (start, code - start);
 
 	if (info)
-		*info = mono_tramp_info_create (g_strdup_printf ("call_filter"), start, code - start, ji, unwind_ops);
+		*info = mono_tramp_info_create ("call_filter", start, code - start, ji, unwind_ops);
 
 	return start;
 }
@@ -338,10 +375,9 @@ mono_amd64_throw_exception (guint64 dummy1, guint64 dummy2, guint64 dummy3, guin
 	}
 
 	if (mono_debug_using_mono_debugger ()) {
-		guint8 buf [16], *code;
+		guint8 buf [16];
 
 		mono_breakpoint_clean_code (NULL, (gpointer)rip, 8, buf, sizeof (buf));
-		code = buf + 8;
 
 		if (buf [3] == 0xe8) {
 			MonoContext ctx_cp = ctx;
@@ -357,7 +393,7 @@ mono_amd64_throw_exception (guint64 dummy1, guint64 dummy2, guint64 dummy3, guin
 	/* adjust eip so that it point into the call instruction */
 	ctx.rip -= 1;
 
-	mono_handle_exception (&ctx, exc, (gpointer)rip, FALSE);
+	mono_handle_exception (&ctx, exc);
 	restore_context (&ctx);
 
 	g_assert_not_reached ();
@@ -497,7 +533,7 @@ get_throw_trampoline (MonoTrampInfo **info, gboolean rethrow, gboolean corlib, g
 		ji = mono_patch_info_list_prepend (ji, code - start, MONO_PATCH_INFO_JIT_ICALL_ADDR, corlib ? "mono_amd64_throw_corlib_exception" : "mono_amd64_throw_exception");
 		amd64_mov_reg_membase (code, AMD64_R11, AMD64_RIP, 0, 8);
 	} else {
-		amd64_mov_reg_imm (code, AMD64_R11, resume_unwind ? (mono_amd64_resume_unwind) : (corlib ? (gpointer)mono_amd64_throw_corlib_exception : (gpointer)mono_amd64_throw_exception));
+		amd64_mov_reg_imm (code, AMD64_R11, resume_unwind ? ((gpointer)mono_amd64_resume_unwind) : (corlib ? (gpointer)mono_amd64_throw_corlib_exception : (gpointer)mono_amd64_throw_exception));
 	}
 	amd64_call_reg (code, AMD64_R11);
 	amd64_breakpoint (code);
@@ -509,7 +545,7 @@ get_throw_trampoline (MonoTrampInfo **info, gboolean rethrow, gboolean corlib, g
 	nacl_global_codeman_validate(&start, kMaxCodeSize, &code);
 
 	if (info)
-		*info = mono_tramp_info_create (g_strdup (tramp_name), start, code - start, ji, unwind_ops);
+		*info = mono_tramp_info_create (tramp_name, start, code - start, ji, unwind_ops);
 
 	return start;
 }
@@ -569,7 +605,6 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls,
 
 	memset (frame, 0, sizeof (StackFrameInfo));
 	frame->ji = ji;
-	frame->managed = FALSE;
 
 	*new_ctx = *ctx;
 
@@ -581,9 +616,6 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls,
 
 		frame->type = FRAME_TYPE_MANAGED;
 
-		if (!ji->method->wrapper_type || ji->method->wrapper_type == MONO_WRAPPER_DYNAMIC_METHOD)
-			frame->managed = TRUE;
-
 		if (ji->from_aot)
 			unwind_info = mono_aot_get_unwind_info (ji, &unwind_info_len);
 		else
@@ -638,12 +670,8 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls,
 
 #ifndef MONO_AMD64_NO_PUSHES
 		/* Pop arguments off the stack */
-		{
-			MonoJitArgumentInfo *arg_info = g_newa (MonoJitArgumentInfo, mono_method_signature (ji->method)->param_count + 1);
-
-			guint32 stack_to_pop = mono_arch_get_argument_info (mono_method_signature (ji->method), mono_method_signature (ji->method)->param_count, arg_info);
-			new_ctx->rsp += stack_to_pop;
-		}
+		if (ji->has_arch_eh_info)
+			new_ctx->rsp += mono_jit_info_get_arch_eh_info (ji)->stack_size;
 #endif
 
 		return TRUE;
@@ -726,9 +754,9 @@ mono_arch_find_jit_info (MonoDomain *domain, MonoJitTlsData *jit_tls,
  *   Called by resuming from a signal handler.
  */
 static void
-handle_signal_exception (gpointer obj, gboolean test_only)
+handle_signal_exception (gpointer obj)
 {
-	MonoJitTlsData *jit_tls = TlsGetValue (mono_jit_tls_id);
+	MonoJitTlsData *jit_tls = mono_native_tls_get_value (mono_jit_tls_id);
 	MonoContext ctx;
 	static void (*restore_context) (MonoContext *);
 
@@ -740,11 +768,27 @@ handle_signal_exception (gpointer obj, gboolean test_only)
 	if (mono_debugger_handle_exception (&ctx, (MonoObject *)obj))
 		return;
 
-	mono_handle_exception (&ctx, obj, MONO_CONTEXT_GET_IP (&ctx), test_only);
+	mono_handle_exception (&ctx, obj);
 
 	restore_context (&ctx);
 }
 
+void
+mono_arch_setup_async_callback (MonoContext *ctx, void (*async_cb)(void *fun), gpointer user_data)
+{
+	guint64 sp = ctx->rsp;
+
+	ctx->rdi = (guint64)user_data;
+
+	/* Allocate a stack frame below the red zone */
+	sp -= 128;
+	/* The stack should be unaligned */
+	if ((sp % 16) == 0)
+		sp -= 8;
+	ctx->rsp = sp;
+	ctx->rip = (guint64)async_cb;
+}
+
 /**
  * mono_arch_handle_exception:
  *
@@ -752,33 +796,24 @@ handle_signal_exception (gpointer obj, gboolean test_only)
  * @obj: the exception object
  */
 gboolean
-mono_arch_handle_exception (void *sigctx, gpointer obj, gboolean test_only)
+mono_arch_handle_exception (void *sigctx, gpointer obj)
 {
 #if defined(MONO_ARCH_USE_SIGACTION)
-	ucontext_t *ctx = (ucontext_t*)sigctx;
+	MonoContext mctx;
 
 	/*
 	 * Handling the exception in the signal handler is problematic, since the original
 	 * signal is disabled, and we could run arbitrary code though the debugger. So
 	 * resume into the normal stack and do most work there if possible.
 	 */
-	MonoJitTlsData *jit_tls = TlsGetValue (mono_jit_tls_id);
-	guint64 sp = UCONTEXT_REG_RSP (ctx);
+	MonoJitTlsData *jit_tls = mono_native_tls_get_value (mono_jit_tls_id);
 
 	/* Pass the ctx parameter in TLS */
-	mono_arch_sigctx_to_monoctx (ctx, &jit_tls->ex_ctx);
-	/* The others in registers */
-	UCONTEXT_REG_RDI (ctx) = (guint64)obj;
-	UCONTEXT_REG_RSI (ctx) = test_only;
+	mono_arch_sigctx_to_monoctx (sigctx, &jit_tls->ex_ctx);
 
-	/* Allocate a stack frame below the red zone */
-	sp -= 128;
-	/* The stack should be unaligned */
-	if (sp % 8 == 0)
-		sp -= 8;
-	UCONTEXT_REG_RSP (ctx) = sp;
-
-	UCONTEXT_REG_RIP (ctx) = (guint64)handle_signal_exception;
+	mctx = jit_tls->ex_ctx;
+	mono_arch_setup_async_callback (&mctx, handle_signal_exception, obj);
+	mono_monoctx_to_sigctx (&mctx, sigctx);
 
 	return TRUE;
 #else
@@ -789,7 +824,7 @@ mono_arch_handle_exception (void *sigctx, gpointer obj, gboolean test_only)
 	if (mono_debugger_handle_exception (&mctx, (MonoObject *)obj))
 		return TRUE;
 
-	mono_handle_exception (&mctx, obj, MONO_CONTEXT_GET_IP (&mctx), test_only);
+	mono_handle_exception (&mctx, obj);
 
 	mono_arch_monoctx_to_sigctx (&mctx, sigctx);
 
@@ -800,85 +835,13 @@ mono_arch_handle_exception (void *sigctx, gpointer obj, gboolean test_only)
 void
 mono_arch_sigctx_to_monoctx (void *sigctx, MonoContext *mctx)
 {
-#if defined(__native_client_codegen__) || defined(__native_client__)
-	printf("WARNING: mono_arch_sigctx_to_monoctx() called!\n");
-#endif
-
-#if defined(MONO_ARCH_USE_SIGACTION)
-	ucontext_t *ctx = (ucontext_t*)sigctx;
-
-	mctx->rax = UCONTEXT_REG_RAX (ctx);
-	mctx->rbx = UCONTEXT_REG_RBX (ctx);
-	mctx->rcx = UCONTEXT_REG_RCX (ctx);
-	mctx->rdx = UCONTEXT_REG_RDX (ctx);
-	mctx->rbp = UCONTEXT_REG_RBP (ctx);
-	mctx->rsp = UCONTEXT_REG_RSP (ctx);
-	mctx->rsi = UCONTEXT_REG_RSI (ctx);
-	mctx->rdi = UCONTEXT_REG_RDI (ctx);
-	mctx->rip = UCONTEXT_REG_RIP (ctx);
-	mctx->r12 = UCONTEXT_REG_R12 (ctx);
-	mctx->r13 = UCONTEXT_REG_R13 (ctx);
-	mctx->r14 = UCONTEXT_REG_R14 (ctx);
-	mctx->r15 = UCONTEXT_REG_R15 (ctx);
-#else
-	MonoContext *ctx = (MonoContext *)sigctx;
-
-	mctx->rax = ctx->rax;
-	mctx->rbx = ctx->rbx;
-	mctx->rcx = ctx->rcx;
-	mctx->rdx = ctx->rdx;
-	mctx->rbp = ctx->rbp;
-	mctx->rsp = ctx->rsp;
-	mctx->rsi = ctx->rsi;
-	mctx->rdi = ctx->rdi;
-	mctx->rip = ctx->rip;
-	mctx->r12 = ctx->r12;
-	mctx->r13 = ctx->r13;
-	mctx->r14 = ctx->r14;
-	mctx->r15 = ctx->r15;
-#endif
+	mono_sigctx_to_monoctx (sigctx, mctx);
 }
 
 void
 mono_arch_monoctx_to_sigctx (MonoContext *mctx, void *sigctx)
 {
-#if defined(__native_client__) || defined(__native_client_codegen__)
-  printf("WARNING: mono_arch_monoctx_to_sigctx() called!\n");
-#endif
-
-#if defined(MONO_ARCH_USE_SIGACTION)
-	ucontext_t *ctx = (ucontext_t*)sigctx;
-
-	UCONTEXT_REG_RAX (ctx) = mctx->rax;
-	UCONTEXT_REG_RBX (ctx) = mctx->rbx;
-	UCONTEXT_REG_RCX (ctx) = mctx->rcx;
-	UCONTEXT_REG_RDX (ctx) = mctx->rdx;
-	UCONTEXT_REG_RBP (ctx) = mctx->rbp;
-	UCONTEXT_REG_RSP (ctx) = mctx->rsp;
-	UCONTEXT_REG_RSI (ctx) = mctx->rsi;
-	UCONTEXT_REG_RDI (ctx) = mctx->rdi;
-	UCONTEXT_REG_RIP (ctx) = mctx->rip;
-	UCONTEXT_REG_R12 (ctx) = mctx->r12;
-	UCONTEXT_REG_R13 (ctx) = mctx->r13;
-	UCONTEXT_REG_R14 (ctx) = mctx->r14;
-	UCONTEXT_REG_R15 (ctx) = mctx->r15;
-#else
-	MonoContext *ctx = (MonoContext *)sigctx;
-
-	ctx->rax = mctx->rax;
-	ctx->rbx = mctx->rbx;
-	ctx->rcx = mctx->rcx;
-	ctx->rdx = mctx->rdx;
-	ctx->rbp = mctx->rbp;
-	ctx->rsp = mctx->rsp;
-	ctx->rsi = mctx->rsi;
-	ctx->rdi = mctx->rdi;
-	ctx->rip = mctx->rip;
-	ctx->r12 = mctx->r12;
-	ctx->r13 = mctx->r13;
-	ctx->r14 = mctx->r14;
-	ctx->r15 = mctx->r15;
-#endif
+	mono_monoctx_to_sigctx (mctx, sigctx);
 }
 
 gpointer
@@ -897,7 +860,7 @@ mono_arch_ip_from_context (void *sigctx)
 static void
 restore_soft_guard_pages (void)
 {
-	MonoJitTlsData *jit_tls = TlsGetValue (mono_jit_tls_id);
+	MonoJitTlsData *jit_tls = mono_native_tls_get_value (mono_jit_tls_id);
 	if (jit_tls->stack_ovf_guard_base)
 		mono_mprotect (jit_tls->stack_ovf_guard_base, jit_tls->stack_ovf_guard_size, MONO_MMAP_NONE);
 }
@@ -935,7 +898,7 @@ altstack_handle_and_restore (void *sigctx, gpointer obj, gboolean stack_ovf)
 		restore_context (&mctx);
 	}
 
-	mono_handle_exception (&mctx, obj, MONO_CONTEXT_GET_IP (&mctx), FALSE);
+	mono_handle_exception (&mctx, obj);
 	if (stack_ovf)
 		prepare_for_guard_pages (&mctx);
 	restore_context (&mctx);
@@ -944,12 +907,13 @@ altstack_handle_and_restore (void *sigctx, gpointer obj, gboolean stack_ovf)
 void
 mono_arch_handle_altstack_exception (void *sigctx, gpointer fault_addr, gboolean stack_ovf)
 {
-#if defined(MONO_ARCH_USE_SIGACTION) && defined(UCONTEXT_GREGS)
+#if defined(MONO_ARCH_USE_SIGACTION)
 	MonoException *exc = NULL;
 	ucontext_t *ctx = (ucontext_t*)sigctx;
 	MonoJitInfo *ji = mini_jit_info_table_find (mono_domain_get (), (gpointer)UCONTEXT_REG_RIP (sigctx), NULL);
 	gpointer *sp;
 	int frame_size;
+	ucontext_t *copied_ctx;
 
 	if (stack_ovf)
 		exc = mono_domain_get ()->stack_overflow_ex;
@@ -965,18 +929,30 @@ mono_arch_handle_altstack_exception (void *sigctx, gpointer fault_addr, gboolean
 	 * 128 is the size of the red zone
 	 */
 	frame_size = sizeof (ucontext_t) + sizeof (gpointer) * 4 + 128;
+#ifdef __APPLE__
+	frame_size += sizeof (*ctx->uc_mcontext);
+#endif
 	frame_size += 15;
 	frame_size &= ~15;
 	sp = (gpointer)(UCONTEXT_REG_RSP (sigctx) & ~15);
 	sp = (gpointer)((char*)sp - frame_size);
+	copied_ctx = (ucontext_t*)(sp + 4);
 	/* the arguments must be aligned */
 	sp [-1] = (gpointer)UCONTEXT_REG_RIP (sigctx);
 	/* may need to adjust pointers in the new struct copy, depending on the OS */
-	memcpy (sp + 4, ctx, sizeof (ucontext_t));
+	memcpy (copied_ctx, ctx, sizeof (ucontext_t));
+#ifdef __APPLE__
+	{
+		guint8 * copied_mcontext = (guint8*)copied_ctx + sizeof (ucontext_t);
+		/* uc_mcontext is a pointer, so make a copy which is stored after the ctx */
+		memcpy (copied_mcontext, ctx->uc_mcontext, sizeof (*ctx->uc_mcontext));
+		copied_ctx->uc_mcontext = (void*)copied_mcontext;
+	}
+#endif
 	/* at the return form the signal handler execution starts in altstack_handle_and_restore() */
 	UCONTEXT_REG_RIP (sigctx) = (unsigned long)altstack_handle_and_restore;
 	UCONTEXT_REG_RSP (sigctx) = (unsigned long)(sp - 1);
-	UCONTEXT_REG_RDI (sigctx) = (unsigned long)(sp + 4);
+	UCONTEXT_REG_RDI (sigctx) = (unsigned long)(copied_ctx);
 	UCONTEXT_REG_RSI (sigctx) = (guint64)exc;
 	UCONTEXT_REG_RDX (sigctx) = stack_ovf;
 #endif
@@ -1103,7 +1079,7 @@ mono_arch_get_throw_pending_exception (MonoTrampInfo **info, gboolean aot)
 	nacl_global_codeman_validate(&start, kMaxCodeSize, &code);
 
 	if (info)
-		*info = mono_tramp_info_create (g_strdup_printf ("throw_pending_exception"), start, code - start, ji, unwind_ops);
+		*info = mono_tramp_info_create ("throw_pending_exception", start, code - start, ji, unwind_ops);
 
 	return start;
 }
